package algorithm;

import instance.*;
import instance.FixedMatch.Added;

import java.util.*;

import main.MCSP;

public class SampleGraph {
	Instance inst;
	int k;
	boolean isNoInstance;
	Set<Marker> V1, V2, V;
	Set<Edge> blackEdges;
	List<Marker> blackVertices;
	HashMap<Marker,Set<SampleGraph.Edge>> incidentEdges;
	enum Color{black,green,red,blue,orange};
	class Edge{
		Color color;
		Marker v1,v2;
		boolean forward;
		
		Edge(Color c, Marker v1, Marker v2, boolean f){
			this.color = c;
			this.v1 = v1;
			this.v2 = v2;
			this.forward = f;
		}
		
		public String toString(){
			return (" "+v1.tag()+"-"+v2.tag()+" "+color);
		}
	}
	
	public SampleGraph(int k) {
		this.V1=new HashSet<Marker>();
		this.V2=new HashSet<Marker>();
		this.V=new HashSet<Marker>();
		this.blackEdges = new HashSet<SampleGraph.Edge>();
		this.blackVertices = new LinkedList<Marker>();
		incidentEdges = new HashMap<Marker,Set<SampleGraph.Edge>>();
		this.k = k;
		this.isNoInstance = false;
	}
	
	public boolean hasBlackParallel(){
		boolean parallel = false;
		ListIterator<Marker> it = blackVertices.listIterator();		
		while( it.hasNext()&&!parallel) {
			Marker m = it.next();
			ListIterator<Marker> it2 = blackVertices.listIterator(blackVertices.indexOf(m));		
			while (it2.hasNext()&&!parallel){
				Marker m2 = it2.next();
				if (inst.checkParallel(m, m2)) parallel = true;
			}
		} 
		return parallel;
	}
	
	SampleGraph(Instance inst){
		this(inst.getK());
		MCSP.print("Building Sample Graph");
		this.inst = inst;
		Genome genome1 = inst.getGenome1();
		Genome genome2 = inst.getGenome2();
		//initialize vertex sets
		addGenomeMarkers(genome1,V1);
		addGenomeMarkers(genome2,V2);
		//initialize edges
		for (Iterator<Marker> it = V.iterator();it.hasNext();){
			incidentEdges.put(it.next(), new HashSet<SampleGraph.Edge>());
		}
		//add one color at a time
		MCSP.print("Adding black edges");
		addBlackEdges();
		addGreenEdges(genome1,genome2);
		addRedEdges(genome1,genome2);		
	}
		
	private void addGreenEdges(Genome genome1, Genome genome2) {
		MCSP.printV("Trying green for black edge");		
		for (Iterator<Edge> it = blackEdges.iterator();it.hasNext();){
			Edge be = it.next();
			if (be.forward){ //green only for forward matches
				MCSP.printV("Forward black edge "+be.toString());
				Iterator<Marker> mIt = genome1.startFromMarker(be.v1);
				Iterator<Marker> mIt2 = genome2.startFromMarker(be.v2);
				Marker point = mIt.next();
				Marker point2 = mIt2.next();
				while (mIt.hasNext()&&mIt2.hasNext()&&point.isForwardMatch(point2)){
					point = mIt.next();
					point2 = mIt2.next();
					if (point.isForwardMatch(point2)&&!(blackVertices.contains(point))&&!(blackVertices.contains(point2))){
						Edge ge = new Edge(Color.green,point,point2,true);
						addEdge(point,point2,ge);
						MCSP.printV("Added green edge" + ge.toString());
					}
				}
			}
		}	
	}

	private void addRedEdges(Genome genome1, Genome genome2) {
		MCSP.print("Adding red edges for black edge");
		for (Iterator<Edge> it = blackEdges.iterator();it.hasNext();){
			Edge be = it.next();
			if (be.forward){ //red only for forward matches
				ListIterator<Marker> mIt = genome1.startFromMarker(be.v1);
				ListIterator<Marker> mIt2 = genome2.startFromMarker(be.v2);
				Marker point = be.v1;
				Marker point2 = be.v2;
				while (mIt.hasPrevious()&&mIt2.hasPrevious()&&point.isForwardMatch(point2)){
					point = mIt.previous();
					point2 = mIt2.previous();
					if (point.isForwardMatch(point2)&&!(blackVertices.contains(point))&&!(blackVertices.contains(point2))){
						Edge re = new Edge(Color.red,point,point2,true);
						addEdge(point, point2,re);
						MCSP.print("Added red edge "+re.toString());
					}
				}
			}
		}	
	}
	
	private void addBlackEdges(){
		//one endpoint must be in V1
		for(Iterator<Marker> it = V1.iterator();it.hasNext()&&!isNoInstance;){
			Marker m = it.next();
			if (m.isFixed()){
				Marker m2 = m.getMatch();
				//both markers must be fixed
				if (m2.isFixed()){
					Edge be =new Edge(Color.black,m,m2,m.isForwardMatch(m2));
					addEdge(m,m2,be);
					MCSP.printV("Added black edge"+ be.toString());
					blackEdges.add(be);
					if (blackEdges.size()>k) isNoInstance = true;
					blackVertices.add(m);
					blackVertices.add(m2);
				}
			}
		}
		MCSP.print("Number of black edges: " + blackEdges.size());
		//inst.print();
	}
	
	private void addGenomeMarkers(Genome g, Set<Marker> s){
		for(ListIterator<Marker> it = g.iterator();it.hasNext();){
			Marker m = it.next();
			s.add(m);
			V.add(m);
		}
	}
	
	private void addEdge(Marker v1, Marker v2,Edge e){
		this.incidentEdges.get(v1).add(e);
		this.incidentEdges.get(v2).add(e);
	}

	private int degree(Marker m){
		if (this.incidentEdges.containsKey(m))	return this.incidentEdges.get(m).size();
		else return 0;
	}	
	
	public Marker findIsolated(){
		Marker minDeg=null;
		for (Iterator<Marker> it = V.iterator(); it.hasNext();) {
			Marker m = it.next();
			if (degree(m)==0&&m.isRare()){
				if (minDeg==null){
					minDeg=m;
				}
				else if (degree(m)==0&&minDeg.matchNum()>m.matchNum()&&m.isRare()){
					minDeg=m;
				}
			}
		}		
		if (minDeg!=null) MCSP.print("Minimum Match Number of isolated: " + minDeg.matchNum());
		else MCSP.print("No isolated");
		return minDeg;
	}

	private Set<Marker> getNeighbors(Marker v){
		if (!V.contains(v)||!incidentEdges.containsKey(v)) return null;
		else {
			Set<Marker> neighbors = new HashSet<Marker>();
			for (Edge e : incidentEdges.get(v)){
				neighbors.add(e.v1);
				neighbors.add(e.v2);
			}
			neighbors.remove(v);
			return neighbors;
		}
	}
	
	public Set<Marker> findOddPath(){
		MCSP.print("Looking for rare odd paths");
		Set<Marker> candidatePath = null;
		for (Marker v : V) {
			//find path starting on v, v must be rare
			if (v.isRare() && degree(v)==1&&!blackVertices.contains(v)){//only need to look for rare odd paths
				MCSP.print("Starting path at "+v.tag());
				Set<Marker> currentPath = new HashSet<Marker>();
				currentPath.add(v);
				Set<Marker> neighbors = getNeighbors(v);
				for (Marker m: neighbors){
					if (!currentPath.contains(m)){//find growing end
						v=m;
						MCSP.print("Growing path with "+v.tag());
					}
				}
				currentPath.addAll(neighbors);	
				boolean mayGrow=true;
				boolean isCycle=false;
				while (mayGrow){
					if (degree(v)==1){
						MCSP.print("Reached End Of Path");
						mayGrow=false;
					}
					else {
						neighbors = getNeighbors(v);
						boolean foundNew=false;
						for (Marker m: neighbors){
							if (!currentPath.contains(m)){//find growing end
								v=m;
								foundNew = true;
								MCSP.print("Growing path with "+v.name);
							}
						}
						if (!foundNew){
							isCycle = true;
							MCSP.print("Is a cycle");
							mayGrow = false;
						}
						currentPath.addAll(neighbors);				
					}
				}
				MCSP.print("Parity of path: " +(currentPath.size() % 2));
				if (!isCycle&&!((currentPath.size() % 2) == 0)){
					if (candidatePath==null||
						candidatePath.size()*candidatePath.iterator().next().matchNum()>
					    currentPath.size()*v.matchNum()) 
						candidatePath=currentPath;
					MCSP.print("Found odd path");
					}				
			}
		}
		
		MCSP.print("Have not found odd path");
		return candidatePath;
		
	}
	
	public void addFixes(){		
		for (Iterator<Edge> it = blackEdges.iterator();it.hasNext();){
			Edge edge = it.next();
			fixEdge(edge);
			it.remove();
		}
		//checkForDegreeOneGuys
		boolean hasDegreeOne=true;
		while (hasDegreeOne){
			hasDegreeOne=false;
			for (Marker v: V){
				if (degree(v)==1&&!blackVertices.contains(v)){
					hasDegreeOne = true;
					//degree one and rare=>even path either red or green
					if (v.isRare()){
						Edge edge = getEdge(v);
						fixAndIsolateEdge(edge);
					}
					//degree one and abundant=>odd path take green edge
					else {
						Edge edge = getEdge(v);
						if (edge.color==Color.green){
							fixAndIsolateEdge(edge);
						}
					}
					MCSP.printV("Found Degree One");
				}
			}
		}
		MCSP.printV("Removed all paths");		
		for (Marker v: V){
			if (degree(v)==2){
				Iterator<Edge> it = incidentEdges.get(v).iterator();
				Edge edge1=it.next();
				Edge edge2=it.next();
				if (edge1.color==Color.green){
					fixAndIsolateEdge(edge1);
					MCSP.printV("recursing");
					//recurse to find paths or other degree two
					addFixes();
				}
				else if (edge2.color==Color.green){
					fixAndIsolateEdge(edge2);
					//recurse to find paths or other degree two
					MCSP.printV("recursing");
					addFixes();
				}
				else {
					System.out.println("A vertex with two non-green edges should not exist");
				}				
			}
		}
	}
	
	private Edge getEdge(Marker m){
		if (this.incidentEdges.containsKey(m)) 
			return this.incidentEdges.get(m).iterator().next();
		else return null;
	}
	
	private void fixEdge(Edge e){		
		inst.getTempSol().addFix(e.v1, e.v2,e.color==Color.black,Added.End);
	}
	
	
	private void fixAndIsolateEdge(Edge e){
		inst.getTempSol().addFix(e.v1, e.v2,e.color==Color.black,Added.End);
		makeIsolated(e.v1);
		makeIsolated(e.v2);
	}
	
	private void makeIsolated(Marker m){
		Set<Marker> Nm = getNeighbors(m);
		incidentEdges.remove(m);
		for (Marker n:Nm){
			for (Iterator<Edge> it= incidentEdges.get(n).iterator(); it.hasNext();) {
				Edge e=  it.next();
				if (e.v1==m||e.v2==m) it.remove();
			} 
		}
	}
	
}